package CleanCSS;
use warnings;
use strict;

use lib './third-party/HTML-Tree-3.23/lib';
use lib './third-party/CSS-SAC-0.08/blib/lib';
use lib './third-party/Class-ArrayObjects-1.03/lib';

use base 'Exporter';

our @EXPORT = qw(clean_css clean_css_decl);

use CSS::SAC;
use CSS::SAC::Condition ':constants';
use CSS::SAC::Selector ':constants';
use CSS::SAC::LexicalUnit ':constants';

# List of properties for the CSS2.1 spec, appendix F. This is not where filtering
# for URLs, etc. is done, and we assume except for URLs all properties in the spec
# are safe to pass through to browsers. (Some filtering of property names is
# necessary to keep out, e.g., 'expression'.)
my @PROPERTIES = qw/
    align
    azimuth
    background-attachment
    background-color
    background-position
    background-repeat
    background-image
    background
    border-collapse
    border-color
    border-spacing
    border-style
    border-top border-right border-left border-bottom
    border-top-color border-right-color border-left-color border-bottom-color
    border-top-style border-right-style border-left-style border-bottom-style
    border-top-width border-right-width border-left-width border-bottom-width
    border-width
    border
    bottom
    caption-side
    clear
    clip
    color
    content
    counter-increment
    counter-reset
    cue-after
    cue-before
    cue
    cursor
    direction
    display
    elevation
    empty-cells
    float
    font-family font-size font-style font-variant font-weight font
    height
    left
    letter-spacing
    line-height
    list-tyle-image list-style-position list-style-type list-style
    margin-right margin-left margin-top margin-bottom margin
    max-height max-width
    min-height min-width
    orphans
    outline-color
    outline-style
    outline-width
    outline
    overflow
    padding-top padding-right padding-bottom padding-left padding
    page-break-after page-break-before page-break-inside
    pause-after pause-before pause
    pitch-range pitch
    play-during
    position
    quotes
    richness
    right
    speak-header speak-numeral speak-punctuation speak speech-rate stress
    table-layout text-align text-decoration text-indent text-transform
    top
    unicode-bidi
    vertical-align
    visibility
    voice-family volume
    white-space
    widows
    width
    word-spacing
    z-index
/;

my %PROPERTIES = map { $_ => 1 } @PROPERTIES;

sub _abort_unit {
    die "ABORTED\n";
}


# Lexical unit handling.

sub _resolve_unit {
    my ($lu) = @_;
    my ($value) = $lu->Value() =~ /^([0-9]+(?:\.[0-9]*)?|\.[0-9]+)\z/;
    if (not defined $value) { $value = 0 }
    my ($unit_type) = $lu->DimensionUnitText =~ /^([a-z]*|%|)\z/;
    $unit_type ||= "";
    return "$value$unit_type";
}

sub _resolve_function {
    my ($lu) = @_;
    my ($value) = $lu->Value() =~ /^((?:\s*[-_a-zA-Z0-9]+\s*,?)+(?<!,))\z/;
    if (not defined $value) { _abort_unit() }
    my ($func_name) = $lu->FunctionName =~ /^([a-z]+)\z/;
    return "$func_name($value)";
}

sub _resolve_simple {
    my ($lu) = @_;
    my ($value) = $lu->Value() =~ /^([-+_a-zA-Z0-9]+(?:\.[0-9]*)?|[0-9]*\.[0-9]+)\z/;
    if (not defined $value) { _abort_unit() }
    return $value;
}

sub _resolve_unicode_range {
    my ($lu) = @_;
    my ($value) = $lu->Value() =~ 
        /^(U\+[0-9a-fA-F\?]{1,6}|U\+[0-9a-fA-F]{1,6}\-[0-9a-fA-F]{1,6})\z/;
    if (!$value) { _abort_unit() }
    return $value;
}

# The CSS string syntax is considerably more permissive.
# We disallow nested quotes in strings entirely (though they should be
# escapable with since some old browsers apparently don't do this (and
# we don't want to have to consider how both parses will work).
sub _format_string {
    my ($value) = @_;
    my $ESCAPE_RE = qr/\\[0-9a-fA-F]{1,6}/;
    # printable ASCII except vertical whitespace and \ (except for \HHHHHH) and
    # either ' or "
    ($value) = $value =~ /^(
            (?:[\t -&(-\[\]-~]|$ESCAPE_RE)*|
            (?:[\t -!#-\[\]-~]|$ESCAPE_RE)*
        )\z/x;
    if (not defined $value) { _abort_unit() }
    if ($value !~ /'/) {
        return qq/'$value'/;
    } elsif ($value !~ /"/) {
        return qq/"$value"/;
    } else {
        _abort_unit();
    }
}

sub _resolve_string {
    my ($lu) = @_;
    return _format_string($lu->Value());
}

sub _resolve_rgbcolor {
    my ($lu) = @_;
    if ($lu->FunctionName eq 'rgb') {
        return _resolve_function($lu);
    } else {
        my ($value) = $lu->Value() =~ /([0-9a-fA-F]{3}|[0-9a-fA-F]{6})/;
        if (defined $value) {
            return "#$value";
        } else {
            _abort_unit();
        }
    }
}

# Lexical unit handlers; entries not in this table should call _abort_unit.
my %LU_RESOLVE = (
    (map { $_ => \&_resolve_unit } 
        CENTIMETER, DEGREE, DIMENSION, EM, EX, GRADIAN, HERTZ, INCH, KILOHERTZ, MILLIMETER,
        MILLISECOND, PERCENTAGE, PICA, PIXEL, POINT, RADIAN, SECOND),
    URI() => \&_abort_unit,
    ATTR() => \&_abort_unit, # might be used to specify URIs
    FUNCTION() => \&_abort_unit, 
    (map { $_ => \&_resolve_function }
        COUNTER_FUNCTION, COUNTERS_FUNCTION, RECT_FUNCTION),
    INHERIT() => sub { 'inherit' },
    (map { $_ => \&_resolve_simple } IDENT, INTEGER, REAL),
    UNICODERANGE() => \&_resolve_unicode_range,
    STRING_VALUE() => \&_resolve_string,
    RGBCOLOR() => \&_resolve_rgbcolor,
);

sub _resolve_lu {
    my ($lu) = @_;
    if ($lu->is_type(OPERATOR_COMMA)) {
        return ',';
    } elsif ($lu->is_type(OPERATOR_SLASH)) {
        return '/';
    } elsif (exists $LU_RESOLVE{$lu->LexicalUnitType()}) {
        return $LU_RESOLVE{$lu->LexicalUnitType()}->($lu);
    } else {
        _abort_unit();
    }
}

# Condition handling
sub _resolve_attribute_condition {
    # we ignore namespace handling.
    my ($cond) = @_;
    my $result = '[';
    if (defined $cond->LocalName) {
        my ($lName) = $cond->LocalName() =~ /([a-zA-Z0-9][-_a-zA-Z0-9]+)/;
        if (not defined $lName) { _abort_unit(); }
        $result .= $lName;
    } else {
        $result .= '*';
    }

    my %OPS = (
        ATTRIBUTE_CONDITION() => '=',
        BEGIN_HYPHEN_ATTRIBUTE_CONDITION() => '|=',
        ONE_OF_ATTRIBUTE_CONDITION() => '~=',
        STARTS_WITH_ATTRIBUTE_CONDITION() => '^=',
        ENDS_WITH_ATTRIBUTE_CONDITION() => '$=',
        CONTAINS_ATTRIBUTE_CONDITION() => '*=',
    );

    if ($cond->Specified) {
        my $op = $OPS{$cond->ConditionType()} || '=';
        $result .= $op;
        $result .= _format_string($cond->Value);
    }

    $result .= ']';
    return $result;
}

sub _resolve_class_condition {
    my ($cond) = @_;
    my ($value) = $cond->Value =~ /([a-zA-Z][-_a-zA-Z0-9]*)/;
    return ".$value";
}

sub _resolve_contains_condition {
    my ($cond) = @_;
    return ':contains(' . _format_string($cond->Value) . ')';
}

sub _resolve_id_condition {
    my ($cond) = @_;
    my ($value) = $cond->Value =~ /([a-zA-Z][-_a-zA-Z0-9]*)/;
    if (not defined $value) { _abort_unit() }
    return "#$value";
}

sub _resolve_lang_condition {
    my ($cond) = @_;
    my ($value) = $cond->Lang =~ /([a-zA-Z][-_a-zA-Z0-9]*)/;
    if (not defined $value) { _abort_unit() }
    return ":lang($value)";
}

# Pseudoclasses and pseudoelements from CSS2.1 section 5.11. :visited is
# deliberately omitted.
my @PSEUDOCLASSES = qw/
    first-child
    hover
    active
    focus
    link
    before
    after
/;

my %PSEUDOCLASSES = map { $_ => 1 } @PSEUDOCLASSES;

sub _resolve_pseudo_class_condition {
    my ($cond) = @_;
    if ($PSEUDOCLASSES{$cond->Value()}) {
        return ':' . $cond->Value();
    } else {
        _abort_unit();
    }
}

sub _resolve_positional {
    my ($cond) = @_;
    my $base = $cond->Type() ? 'of-type' : 'child';
    if ($cond->Position() == 1) {
        return ":first-$base";
    } elsif ($cond->Position() == -1) {
        return ":last-$base";
    } else {
        my ($number) = $cond->Position() =~ /^(-?[0-9]+)\z/;
        if (not defined $number) { _abort_unit() }
        return ":nth-$base($number)";
    }
}


my %COND_RESOLVE = (
   AND_CONDITION() => sub {
           _resolve_selector_or_condition($_[0]->FirstCondition()) .
           _resolve_selector_or_condition($_[0]->SecondCondition())
       },
   (map { $_ => \&_resolve_attribute_condition } 
        ATTRIBUTE_CONDITION, BEGIN_HYPHEN_ATTRIBUTE_CONDITION,
        ONE_OF_ATTRIBUTE_CONDITION, STARTS_WITH_ATTRIBUTE_CONDITION,
        ENDS_WITH_ATTRIBUTE_CONDITION, CONTAINS_ATTRIBUTE_CONDITION),
   CLASS_CONDITION() => \&_resolve_class_condition,
   CONTENT_CONDITION() => \&_resolve_contains_condition,
   ID_CONDITION() => \&_resolve_id_condition,
   LANG_CONDITION() => \&_resolve_lang_condition,
   NEGATIVE_CONDITION() => sub { ':not(' . _resolve_selector_or_condition($_[0]->Condition) . ')' },
   ONLY_CHILD_CONDITION() => sub { ':only-child' },
   ONLY_TYPE_CONDITION() => sub { ':only-of-type' },
   IS_ROOT_CONDITION() => sub { ':root' },
   IS_EMPTY_CONDITION() => sub { ':empty' },
   PSEUDO_CLASS_CONDITION() => \&_resolve_pseudo_class_condition,
   POSITIONAL_CONDITION() => \&_resolve_positional,
);

sub _resolve_selector_or_condition {
    my ($cond) = @_;
    if ($cond->isa('CSS::SAC::Condition')) {
        return _resolve_condition($cond);
    } else {
        return _resolve_selector($cond);
    }
}

sub _resolve_condition {
    my ($cond) = @_;
    if ($cond->isa('CSS::SAC::Condition') and exists $COND_RESOLVE{$cond->ConditionType}) {
        return $COND_RESOLVE{$cond->ConditionType}->($cond);
    } else {
        _abort_unit();
    }
}

sub _combine_selectors {
    my ($first, $middle, $second) = @_;
    $first .= 'Selector';
    $second .= 'Selector';
    return sub {
        my ($selector) = @_;
        return _resolve_selector($selector->$first) . $middle .
               _resolve_selector($selector->$second);
    };
}

sub _resolve_element_node_selector {
    my ($selector) = @_;
    # Namespaces ignored.
    if (defined $selector->LocalName) {
        my ($lName) = $selector->LocalName =~ /^([a-zA-Z][-_a-zA-Z0-9]*)\z/;
        if (not defined $lName) { _abort_unit() }
        return $lName;
    } else {
        return '*';
    }
}

my %SELECTOR_RESOLVE = (
    CHILD_SELECTOR() => _combine_selectors('Ancestor', ' > ', 'Simple'),
    DESCENDANT_SELECTOR() => _combine_selectors('Ancestor', ' ', 'Simple'),
    DIRECT_ADJACENT_SELECTOR() => _combine_selectors('', ' + ', 'Sibling'),
    INDIRECT_ADJACENT_SELECTOR() => _combine_selectors('', ' ~ ', 'Sibling'),
    CONDITIONAL_SELECTOR() => sub {
        _resolve_selector($_[0]->SimpleSelector) . _resolve_selector_or_condition($_[0]->Condition)
    },
    NEGATIVE_SELECTOR() => sub { ':not(' . _resolve_selector($_[0]->SimpleSelector) . ')' },
    ELEMENT_NODE_SELECTOR() => \&_resolve_element_node_selector,
    # pseudo-elements not handled
);

sub _resolve_selector {
    my ($selector) = @_;
    if (exists $SELECTOR_RESOLVE{$selector->SelectorType()}) {
        return $SELECTOR_RESOLVE{$selector->SelectorType()}->($selector);
    } else {
        _abort_unit();
    }
}

# State machine for CSS::SAC
use constant WAIT_FOR_SELECTOR => 0;
use constant IN_RULE => 1;

sub new {
    my $class = shift;
    my $self = { 
        state => WAIT_FOR_SELECTOR,
        output => '',
        level => 0,
    };
    return bless $self, $class;
}

sub start_document {}
sub end_document {
    my ($self) = @_;
    if ($self->{state} == IN_RULE) {
        $self->{output} .= " } ";
        $self->{state} = WAIT_FOR_SELECTOR;
    }
    while ($self->{level}) {
        $self->{output} .= " } ";
        --$self->{level};
    }

}

sub start_selector {
    my ($self, $selectors) = @_;
    if ($self->{state} != WAIT_FOR_SELECTOR) {
        warn "Inconsistent parse (selector started in WAIT_FOR_SELECTOR)\n";
    }
    eval {
        _abort_unit() unless @$selectors;
        my $selector_string = join ', ', map _resolve_selector($_), @$selectors;
        $self->{output} .= " $selector_string {";
        $self->{state} = IN_RULE;
    };
    if ($@ && $@ ne "ABORTED\n") {
        die $@;
    }
}

sub end_selector {
    my ($self) = @_;
    if ($self->{state} == IN_RULE) {
        $self->{output} .= " } ";
        $self->{state} = WAIT_FOR_SELECTOR;
    }
}

sub property {
    my ($self, $name, $units, $important) = @_;
    if ($self->{state} != IN_RULE) {
        return;
    }
    eval {
        if (!$PROPERTIES{$name}) {
            _abort_unit();
        }
        my @resolved_units = map _resolve_lu($_), @$units;
        my $important_string = $important ? ' ! important' : ' ';
        $self->{output} .= " $name: " . join(" ", @resolved_units) . "$important_string;";
    };
    if ($@ && $@ ne "ABORTED\n") {
        die $@;
    } elsif ($@ && $@ eq "ABORTED\n") {
        # warn "Ignoring property $name\n";
    }
}

sub ignorable_at_rule {}

sub import_style {} # deliberately ignored

sub namespace_declaration {} # no namespace support implemented

sub start_media {
    my ($self, $media) = @_;
    my @filtered_media = grep defined, map { /([a-zA-Z][-_a-zA-Z0-9]*)/ } @$media;
    if (@filtered_media) {
        $self->{output} .= " \@media " . join(", ", @filtered_media) . " { ";
        $self->{level}++;
    }
}

sub _end_generic {
    my ($self) = @_;
    if ($self->{level}) {
        $self->{level}--;
        $self->{output} .= " } ";
    }
}

*end_media = \&_end_generic;

sub comment {}

# not implemented
sub charset {}
sub start_font_face {}
sub end_font_face {}
sub start_page {}
sub end_page {}

sub warning {}
sub error {}
sub fatal_error {
    die "Aborted\n";
}


# External interface:
sub clean_css_decl {
    my ($raw_css_rule) = @_;
    my $object = CleanCSS->new;
    $object->{state} = IN_RULE;
    my $sac = CSS::SAC->new({
        DocumentHandler => $object,
        ErrorHandler => $object,
    });
    eval {
        $sac->parse_style_declaration(\$raw_css_rule);
    };
    die $@ if $@ and $@ ne "Aborted\n";
    return $object->{output};
}

sub clean_css {
    my ($raw_css) = @_;
    my $object = CleanCSS->new;
    my $sac = CSS::SAC->new({
        DocumentHandler => $object,
        ErrorHandler => $object,
    });
    eval {
        $sac->parse({
            string => $raw_css,
            embedded => 1,
        });
    };
    die $@ if $@ and $@ ne "Aborted\n";
    if ($@ eq "Aborted\n") {
        return '';
    } else {
        return $object->{output};
    }
}

1;
