package CleanHTML;
use warnings;
use strict;

use base 'Exporter';

our @EXPORT = qw( print_clean_html clean_html );

use HTML::TreeBuilder;
use HTML::HeadParser;
use HTML::Tagset (); # for %HTML::Tagset::emptyElement

use CleanCSS;

# Manaully filtered from whatwg.org/html5 (with some attributes from w3.org/TR/html4)
#   [Many of these aren't actually implemented in current browsers...]
# Unfortunate things to omit:
#   Privacy violation (accessing external web pages without user intervention):
#       link, img, style, audio, video, source, object, embed, iframe, meta
#           (data: URLs could probably be allowed for some content-types, except
#            for considerations concerning vulnerabilities in image/audio/video
#            decoders, etc. For style sheets, we would want to parse and rebuild the
#            the entire CSS segment to avoid quirky browser parsing. Some rel/rev
#            values for link are probably okay, and many types of meta tags can't
#            link traversal or encoding confusion.)
#   Not useful without scripting:
#       canvas
#   Not useful without images:
#       map, area
#   Requires URL filtering to support:
#       base

# Tags that are allowed in output.
my @SAFE_TAGS = qw/
    html head title 
    body section nav article h1 h2 h3 h4 h5 h6
    hgroup header footer address
    p hr br pre dialog blockquote ol ul li dl dt dd
    a em strong small cite q dfn abbr code var samp kbd sub sup
    i b mark progress meter time ruby rt rp bdo span
    ins del
    figure
    table caption colgroup col tbody thead tfoot tr td th
    form fieldset label input button select datalist optgroup option
    textarea keygen output
    legend div

    acronym dir 

    big blink center font s strike tt u
/;

my %IS_SAFE_TAG = map { $_ => 1 } @SAFE_TAGS;

my @SAFE_ATTRS = qw/
    title summary item href cite list accept-charset action name value 
/;

# Tags which get special replacement:
my %SPECIAL_TAGS = (
    'img' => sub { defined $_[0]->attr('alt') ? $_[0]->attr('alt') : '[image]'; },
    'script' => sub { '' },
    'style' => sub {
        $_[0]->tag('notstyle'); # hack around as_text exception
        my $text = $_[0]->as_text();
        $_[0]->tag('style');
        $_[0]->delete_content;
        $_[0]->push_content(clean_css($text));
        return $_[0];
    },
);

sub _matching {
    my ($regex) = @_;
    return sub { if ($_[0] =~ $regex) { return $1 } else { return undef } };
}

sub _nonnegative {
    if ($_[0] =~ /\A\s*\+?([0-9]+)/) {
        return $1
    } else {
        return undef;
    }
}

sub _integer {
    if ($_[0] =~ /\A\s*([-+]?[0-9]+)/) {
        return $1;
    } else {
        return undef;
    }
}

sub _real {
    if ($_[0] =~ /\A\s*([-+]?(?:[0-9]+\.?[0-9]*|\.[0-9]+))/) {
        return $1;
    } else {
        return undef;
    }
}

sub _id {
    # this might be too restrictive for 'id'
    if ($_[0] =~ /\A\s*([-_a-zA-Z0-9]+)\z/) {
        return $1;
    } else {
        return undef;
    }
}

sub _datetime {
    # this doesn't actually validate the datetime format
    if ($_[0] =~ /\A\s*([-+0-9TZ:]{1,80})\z/) {
        return $1; 
    } else {
        return undef;
    }
}

sub _booleans {
    map { my $name = $_; $name => sub { $name } } @_
}

sub _length {
    if ($_[0] =~ /\A\s*([-+]?[0-9]+%?)/) {
        return $1;
    } else {
        return undef;
    }
}

my %AS_SAFE_ATTR = (
    (map { $_ => sub { $_[0] } } @SAFE_ATTRS),
    'dir' => _matching(qr{\A(ltr|rtl)\z}i),
    'class' => _matching(qr{\A([-_A-Za-z0-9 \t]+)\z}),
    'headers' => _matching(qr{\A([-_A-Za-z0-9 \t]+)\z}),
    'accesskey' => _matching(qr{\A([-_A-Za-z0-9 \t]+)\z}),
    'itemprop' => _matching(qr{\A([-_A-Za-z0-9 \t]+)\z}),
    'style' => \&clean_css_decl,
    'clear' => _matching(qr{(left|right|all|none)}),
    'align' => _matching(qr{(top|middle|bottom|left|right)}),
    'valign' => _matching(qr{(top|middle|bottom|baseline)}),
    'char' => _matching(qr{([ -~])}),
    'frame' => _matching(qr{(void|above|below|hsides|lhs|rhs|vsides|box|border)}),
    'rules' => _matching(qr{(none|groups|rows|cols|all)}),
    (map { $_ => _matching(qr{\A(#[0-9a-fA-F]{6})}) } qw/ bgcolor txt link color /),
    'type' => _matching(qr{\A(text/css|hidden|text|search|tel|url|email|
                          password|datetime|date|month|week|time|datetime-local|
                          number|range|color|checkbox|radio|file|submit|reset|button)\z}ix),
    (map { $_ => \&_nonnegative } qw/ span colspan rowspan width height hspace vspace border charoff /),
    (map { $_ => \&_integer } qw/ start tabindex size /),
    (map { $_ => \&_real } qw/ value max min high low optimum /),
    'scope' => _matching(qr{\A(row|col|rowgroup|colgroup)\z}ix),
    (map { $_ => \&_id } qw/ id form formtarget target lang hreflang for itemfor /),
    (map { $_ => \&_datetime } qw/ datetime pubdate /),
    _booleans(qw/scoped reversed disabled readonly required multiple novalidate formnovalidate
                 hidden open selected autofocus nohref noshade compact checked nowrap /),
    'spellcheck' => _matching(qr/(true|false)/),
    'autocomplete' => _matching(qr/(on|off)/),
    'method' => _matching(qr/(GET|POST|PUT|DELETE)/),
    'formmethod' => _matching(qr/(GET|POST|PUT|DELETE)/),
    'enctype' => _matching(qr{(application/x-www-form-urlencoded|multipart/form-data|text/plain)}i),
    'formenctype' => _matching(qr{(application/x-www-form-urlencoded|multipart/form-data|text/plain)}i),
);

# Minimal string filtering (remove control characters, etc.); this does not insert
# entities where they would be appropriate. (That is done output_element.)
sub clean_string {
# http://www.whatwg.org/specs/web-apps/current-work/multipage/syntax.html#preprocessing-the-input-stream for the list of non-accepted characters
    local ($_) = @_;
    s{\xFEFF}{}g;
    s{(?:[\x{0000}-\x{0008}\x{000E}-\x{001F}\x{007F}-\x{009F}\x{D800}-\x{DFFF}]|
       [\x{000B}\x{FFFE}\x{1FFFE}\x{1FFFF}\x{2FFFE}\x{2FFFF}\x{3FFFE}\x{3FFFF}]|
       [\x{4FFFE}\x{4FFFF}\x{5FFFE}\x{5FFFF}\x{6FFFE}\x{6FFFF}\x{7FFFE}\x{7FFFF}]|
       [\x{8FFFE}\x{8FFFF}\x{9FFFE}\x{AFFFE}\x{AFFFF}\x{BFFFE}\x{BFFFF}\x{CFFFE}]|
       [\x{CFFFF}\x{DFFFE}\x{DFFFF}\x{EFFFE}\x{EFFFF}\x{FFFFE}\x{FFFFF}]|
       [\x{10FFFE}\x{10FFFF}\x{11000}-\x{FFFFFFF}])}{\x{FFFD}}xg;
   return $_;
}

sub fix_safe_element {
    my ($element) = @_;
    if (!ref $element) {
        # text
        return clean_string($element)
    } else {
        # warn "Processing safe ", $element->tag(); 
        my @content = map fix_element($_), $element->detach_content;
        $element->push_content($_) for @content;
        for my $attr ($element->all_external_attr_names) {
            next if $attr =~ /^_/; # internal to HTML::Element, won't be output
            if ($AS_SAFE_ATTR{$attr}) {
                my $safe_version = $AS_SAFE_ATTR{$attr}->($element->attr($attr));
                if (defined $safe_version) {
                    $element->attr($attr, clean_string($safe_version));
                } else {
                    $element->attr($attr, undef);
                }
            } else {
                $element->attr($attr, undef);
            }
        }
        return $element
    }
}

# Takes an HTML::Element or string and returns a list of new HTML::Elements.
# Will delete() anything not returned.
sub fix_element {
    my ($element) = @_;
    if (!ref $element || $IS_SAFE_TAG{ $element->tag() }) {
        return fix_safe_element($element);
    } elsif (exists $SPECIAL_TAGS{ $element->tag() }) {
        # warn "Special ", $element->tag(), "\n";
        my $new_element = $SPECIAL_TAGS{ $element->tag() }->($element);
        if (ref $new_element && $new_element != $element) {
            $element->delete;
        }
        return fix_safe_element($new_element);
    } else {
        # warn "Killing " , $element->tag(), "\n";
        my @content = map fix_element($_), $element->detach_content;
        $element->delete;
        return @content;
    }
}

# assumes appropriate tag/attribute filtering has already taken place.
sub output_element {
    my ($element) = @_;
    if (!ref $element) {
        # replace &, ', ", <, > and non-(printable ASCII) with entities.
        $element =~ s{([&"'<>])}{'&#' . ord($1) . ';'}eg;
        $element =~ s{([^\n\r\t -~])}{'&#' . ord($1) . ';'}eg;
        print $element;
    } else {
        print '<';
        print $element->tag();
        for my $attr ($element->all_external_attr_names()) {
            print ' ';
            die if ref $attr;
            output_element($attr);
            print '="';
            die if ref $element->attr($attr);
            output_element($element->attr($attr));
            print '"';
        }
        print '>';
        if (!$HTML::Tagset::emptyElement{$element->tag()}) {
            output_element($_) for $element->content_list;
            print '</';
            print $element->tag();
            print '>';
        }
    }
}

sub print_clean_html {
    my ($document) = @_;

    my $tree = HTML::TreeBuilder->new;

    $tree->parse($document);
    $tree->eof;

    ($tree) = fix_element($tree);
    if (!$tree->find('head')) {
        $tree->unshift_content(
            HTML::Element->new('head')
        );
    }
    my $head = $tree->find('head');
    # Since we entitify all non-ASCII, our output is us-ascii.
    $head->unshift_content(
        HTML::Element->new('meta',
            'http-equiv' => 'Content-Type',
            'content' => 'text/html; charset=us-ascii'));
    print <<'EOS';
<!DOCTYPE HTML>
EOS
    output_element($tree);
}

sub clean_html {
    my ($orig_html) = @_;
    open my $fh, '>', \my $result;
    my $old_fh = select $fh;
    print_clean_html($orig_html);
    select $old_fh;
    return $result;
}

1;
